#!/usr/bin/python3 -tt

import functools
import os

# The bad design here is using attributes of the object to pass parameters.
# One of the issues with that is thread-safety: If the same object is used in two different threads,
# the attributes could be set the wrong way when the file is written
class BadDesign:
    def __init__(self, basedir):
        self.basedir = basedir

    def write(self):
        out_file = os.path.join(self.out_dir, self.filename)
        with open(out_file, 'w') as f:
            f.write(self.data)
        os.chmod(out_file, self.default_mode)

    def run(self, username):
        self.default_mode = 0o644
        self.out_dir = os.path.join(self.basedir, username)
        self.filename = 'data.txt'
        self.data = '12345\n'

        self.write()


# This design is your standard good design
# default_mode is set by __init__() instead of run() because it's a default value.  The other two
# variables are passed as parameters into the write() method
class StandardGoodDesign:
    def __init__(self, basedir, mode=0o644):
        self.basedir = basedir
        self.default_mode = mode

    def write(self, out_dir, filename, data, mode=None):
        out_file = os.path.join(out_dir, filename)
        with open(out_file, 'w') as f:
            f.write(data)
        os.chmod(out_file, self.default_mode)

    def run(self, username):
        out_dir = os.path.join(self.basedir, username)
        filename = 'data.txt'
        data = '12345\n'

        self.write(out_dir, filename, data)


# In some cases, you may have a function that needs to be called multiple times with the same
# parameters.  When that happens, you can use partial:
class GoodDesignWithPartial:
    def __init__(self, basedir, mode=0o644):
        self.basedir = basedir
        self.default_mode = mode

    def write(self, out_dir, filename, data, mode=None):
        out_file = os.path.join(out_dir, filename)
        with open(out_file, 'w') as f:
            f.write(data)
        os.chmod(out_file, self.default_mode)

    def run(self, username):
        out_dir = os.path.join(self.basedir, username)
        filename1 = 'data.txt'
        filename2 = 'data2.txt'
        filename3 = 'data3.txt'
        write_to_dir = functools.partialmethod(self.write, out_dir)
        data1 = '12345\n'
        data2 = 'abcde\n'
        data3 = 'iiiiiiivv\n'

        write_to_dir(filename1, data1)
        write_to_dir(filename2, data2)
        write_to_dir(filename3, data3)


# In other cases, you may have a value that needs to be used with many different functions.
# When that happens, use a second class:
class FileManipulator:
    def __init__(self, out_dir, mode):
        self.out_dir = out_dir
        self.default_mode = mode

    def write(self, filename, data, mode):
        out_file = os.path.join(self.out_dir, filename)
        with open(out_file, 'w') as f:
            f.write(data)
        os.chmod(out_file, self.default_mode)

    def read(self, filename):
        out_file = os.path.join(self.out_dir, filename)
        with open(out_file, 'r') as f:
            return f.read()


class GoodDesignWithFactory:
    def __init__(self, basedir, mode=0o644):
        self.basedir = basedir
        self.default_mode = mode

    def run(self, username):
        out_dir = os.path.join(self.basedir, username)
        manipulator = FileManipulator(out_dir, self.default_mode)

        data1 = manipulator.read('data.txt')
        data2 = manipulator.read('data2.txt')

        manipulator.write('data.txt', data1 + data2)
